from dataclasses import dataclass
from re import sub
from subprocess import check_output
from textwrap import fill
from typing import Optional, Sequence
from warnings import warn

import click


MAX_LINEWIDTH = 75


@dataclass
class ExitSignal(Exception):
    message: Optional[str] = None


def echo(
    message: str = "",
    newline: bool = True,
    raw: bool = False,
    max_linewidth: int = MAX_LINEWIDTH,
):
    if not raw:
        message = rewrap(message, width=max_linewidth)
    click.echo(message, nl=newline)


def get_stripped_output(cmd: Sequence[str]) -> str:
    output: bytes = check_output(cmd)
    return output.decode().strip()


def print_own_version():
    # fmt: off
    try:
        import puprelease
        version = puprelease.__version__
    except ModuleNotFoundError:
        # puprelease is not installed; Running as a script
        import setuptools_scm
        version = setuptools_scm.get_version(root="..", relative_to=__file__)
    echo(f"This is puprelease version {version}")
    # fmt: on


def print_header(header: str):
    """ Underline the given string and prefix with a blank line. """
    echo()
    echo(header)
    echo("-" * len(header))


@dataclass
class KeyValueTable:
    """
    An asynchronously printed, two column table.
    
    "key_column_width" should be at least as large as the longest key, plus the
    length of the separator.
    """

    key_column_width: int
    total_width: Optional[int] = None
    separator = ": "

    def print_row(self, key: str, value: str):
        key_cell = key + self.separator
        if self.key_column_width < len(key_cell):
            warn(f"key_column_width should be at least {len(key_cell)}")
        if self.total_width is None:
            value_cell = value
        else:
            value_column_width = self.total_width - self.key_column_width
            # Make sure the multiline stirng stays within its cell, by rewrapping
            # and prepending spaces to all but the first line.
            value_cell = rewrap(value, value_column_width).replace(
                "\n", "\n" + " " * self.key_column_width
            )

        echo(key_cell.ljust(self.key_column_width) + value_cell, raw=True)


def rewrap(multiline: str, width=70, **TextWrap_kwargs) -> str:
    """ Linearize string, and wrap the result.
    
    :param multiline:  A multiline string as found in indented source code.
    :param width:  Max number of characters per line in the output.
    :param TextWrap_kwargs:  Keyword arguments passed to textwrap.TextWrap
            (from Python's standard library).
    :return:  A string without superfluous whitespace and with newlines so that
            it prints squarely.
    """
    oneline = linearize(multiline)
    return fill(oneline, width=width, **TextWrap_kwargs)


def linearize(multiline: str) -> str:
    """
    :param multiline:  A multiline string as found in indented source code.
    :return:  A stripped, one-line string. All newlines and multiple
            consecutive whitespace characters are replaced by a single space.
    """
    oneline = sub(r"\s+", " ", multiline)
    return oneline.strip()



from os import getcwd
from os.path import exists

from puprelease.util import (
    ExitSignal,
    KeyValueTable,
    get_stripped_output,
    print_header,
)
from requests import get


table = KeyValueTable(key_column_width=42)


def check_package():
    """ Display existing package versions, inter alia. """
    print_header(f"Inspecting Python package")
    package_name = check_setup_py()
    check_pip_installed(package_name)
    check_pypi(package_name)
    check_git_worktree()


def check_setup_py():
    table.print_row("Working directory", getcwd())
    if not exists("setup.py"):
        raise ExitSignal('Working directory does not contain a "setup.py" file')
    fullname = get_stripped_output(["python", "setup.py", "--fullname"])
    package_name, version = fullname.split("-", maxsplit=1)
    table.print_row("Package name", package_name)
    table.print_row('Version in working dir (via "setup.py")', version)
    return package_name


def check_pip_installed(package_name):
    # Note: using pip's internal API to get this info is (too) convoluted.
    lines = get_stripped_output(("pip", "list")).splitlines()
    try:
        line = next(line for line in lines if line.startswith(package_name))
        _, version, *_ = line.split()
    except StopIteration:
        version = "[not pip-installed]"
    table.print_row("System-wide version, installed with pip", version)


def check_pypi(package_name):
    response = get(f"https://pypi.org/pypi/{package_name}/json")
    if response.ok:
        version = response.json()["info"]["version"]
    elif response.status_code == 404:
        version = "[not yet published on pypi.org]"
    else:
        response.raise_for_status()
    table.print_row("Latest version on pypi.org", version)


def check_git_worktree():
    clean = git_worktree_is_clean()
    if clean:
        status = "Clean (no uncommited changes)"
    else:
        status = "Dirty (uncommited changes)"
    table.print_row("Git working tree status", status)
    if not clean:
        raise ExitSignal(
            """Please commit or stash working tree changes before making a new
            release."""
        )


def git_worktree_is_clean():
    # The output of this command (which is the same command as in
    # "setuptools_scm.git.GitWorkdir.is_dirty") is empty when no uncommited
    # files are present in the git working tree.
    cmd = ("git", "status", "--porcelain", "--untracked-files=no")
    output = get_stripped_output(cmd)
    return not output



import click
from puprelease.check import check_package
from puprelease.release import new_release
from puprelease.util import ExitSignal, echo, print_own_version


@click.command()
def pup():
    """
    This program can be safely stopped at any time: the release steps are
    idempotent.
    """
    try:
        print_own_version()
        check_package()
        new_release()
    except click.Abort:
        # Do not print click's ugly "Aborted!"
        pass
    except ExitSignal as sig:
        if sig.message:
            echo()
            echo(sig.message)
    echo("Exiting")


if __name__ == "__main__":
    pup()



from ast import keyword, parse, walk
from dataclasses import dataclass
from os import getenv
from subprocess import list2cmdline, run
from typing import Optional, Sequence

from click import confirm, prompt
from puprelease.util import (ExitSignal, KeyValueTable, MAX_LINEWIDTH, echo,
                             get_stripped_output, print_header)


table = KeyValueTable(key_column_width=13, total_width=MAX_LINEWIDTH)
PyPI_user = getenv("TWINE_USERNAME")


def new_release():
    print_header("Preparing new release")
    confirm("Did you run testsuite locally?", default=True, abort=True)
    if is_versioned_with_git_tags():
        desired_new_version = prompt("Please enter the new version number")
        message = prompt(
            "Please enter a message to go along with the git tag",
            default=get_last_commit_message(),
        )
        git_tag_command(desired_new_version, message).check_and_run()
        worktree_version = get_worktree_version()
        echo()
        echo(f'Version of package in worktree is now: "{worktree_version}"')
        if worktree_version != desired_new_version:
            echo("This does not match the desired new version")
            echo("Removing added tag and quitting")
            revert_tag_command(desired_new_version).run()
            raise ExitSignal
        push_tag_command.check_and_run()
    create_dists_command.check_and_run()
    publish_command.check_and_run()
    echo("Congrats on the new release")


def is_versioned_with_git_tags() -> bool:
    """
    Is the package versioned using git-tags (i.e. using setuptools_scm), or is
    the version set in "setup.py"?
    """
    # Check whether a function call keyword argument of the type
    # "use_scm_version=" appears somewhere in setup.py. Method is not
    # foolproof; But more than good enough.
    with open("setup.py") as f:
        src = f.read()
    tree = parse(src)
    all_kwargs_in_setup_py = [
        node.arg for node in walk(tree) if type(node) == keyword
    ]
    return "use_scm_version" in all_kwargs_in_setup_py


def get_worktree_version():
    return get_stripped_output(("python", "setup.py", "--version"))


def get_last_commit_message():
    return get_stripped_output(("git", "log", "-1", "--pretty=%B"))


@dataclass
class Command:
    title: str
    args: Sequence[str]
    description: Optional[str] = None

    def check_and_run(self):
        print_header(self.title)
        table.print_row("Command", list2cmdline(self.args))
        if self.description:
            table.print_row("Description", self.description)
        confirm("Execute?", default=True, abort=True)
        self.run()

    def run(self):
        echo(list2cmdline(self.args))
        completed_process = run(self.args)
        retcode = completed_process.returncode
        if retcode == 0:
            echo("Command completed succesfully")
        else:
            echo(f"Command completed with return code {retcode}")


def git_tag_command(new_version, message):
    return Command(
        title="Create tag",
        args=("git", "tag", "-a", new_version, "--message", message),
        description=(
            """Create a git tag on the current commit. (Option "-a" makes an
            "annotated" tag, which includes tagger name, email, date, a custom
            message, etc)."""
        ),
    )


def revert_tag_command(new_version):
    return Command(
        title="Remove newly created tag", args=("git", "tag", "-d", new_version)
    )


push_tag_command = Command(
    title="Push tag",
    args=("git", "push", "--tags"),
    description="Push tag to public source code repository.",
)

create_dists_command = Command(
    title="Create distributions",
    args=("python", "setup.py", "bdist_wheel", "sdist"),
    description=(
        """Create two package distributions in "dist/": a built
            distribution and a source distribution. (The source distribution
            includes the project's top-level directory, and requires running
            "setup.py" on each "pip install" -- i.e. requires a full build. The
            built distribution on the other hand only requires moving files on
            "pip install", and includes only source files from the package
            directory.)"""
    ),
)

publish_command = Command(
    title="Publish release",
    args=("twine", "upload", "dist/*", "--skip-existing"),
    description=(
        f'Upload new release to PyPI, using account "{PyPI_user}".'
        if PyPI_user
        else """Upload new release to PyPI. Login details will be prompted
            for. (Note: these can also be set via the TWINE_USERNAME and
            TWINE_PASSWORD environment variables)."""
    ),
)



from setuptools import find_packages, setup


GITHUB_URL = "https://github.com/tfiers/puprelease"

with open("ReadMe.md", mode="r") as f:
    readme = f.read()

setup(
    name="puprelease",
    description="Publishing a new version of your Python package has never been easier",
    author="Tomas Fiers",
    author_email="tomas.fiers@gmail.com",
    long_description=readme,
    long_description_content_type="text/markdown",
    url=GITHUB_URL,
    project_urls={"Source Code": GITHUB_URL},
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["pup=puprelease.pup:pup"]},
    packages=find_packages(),
    install_requires=["click ~=7.0", "requests ~=2.0"],
    # Get package version from git tags
    setup_requires=["setuptools_scm"],
    use_scm_version={
        "version_scheme": "post-release",
        "local_scheme": "dirty-tag",
    },
)
